"use strict";
exports.__esModule = true;
exports.useTracker = void 0;
var meteor_1 = require("meteor/meteor");
var tracker_1 = require("meteor/tracker");
var react_1 = require("react");
// Warns if data is a Mongo.Cursor or a POJO containing a Mongo.Cursor.
function checkCursor(data) {
    var shouldWarn = false;
    if (Package.mongo && Package.mongo.Mongo && data && typeof data === 'object') {
        if (data instanceof Package.mongo.Mongo.Cursor) {
            shouldWarn = true;
        }
        else if (Object.getPrototypeOf(data) === Object.prototype) {
            Object.keys(data).forEach(function (key) {
                if (data[key] instanceof Package.mongo.Mongo.Cursor) {
                    shouldWarn = true;
                }
            });
        }
    }
    if (shouldWarn) {
        console.warn('Warning: your reactive function is returning a Mongo cursor. '
            + 'This value will not be reactive. You probably want to call '
            + '`.fetch()` on the cursor before returning it.');
    }
}
// Used to create a forceUpdate from useReducer. Forces update by
// incrementing a number whenever the dispatch method is invoked.
var fur = function (x) { return x + 1; };
var useForceUpdate = function () { return react_1.useReducer(fur, 0)[1]; };
var useTrackerNoDeps = function (reactiveFn, skipUpdate) {
    if (skipUpdate === void 0) { skipUpdate = null; }
    var refs = react_1.useRef({
        isMounted: false,
        trackerData: null
    }).current;
    var forceUpdate = useForceUpdate();
    // Without deps, always dispose and recreate the computation with every render.
    if (refs.computation) {
        refs.computation.stop();
        // @ts-ignore This makes TS think ref.computation is "never" set
        delete refs.computation;
    }
    // Use Tracker.nonreactive in case we are inside a Tracker Computation.
    // This can happen if someone calls `ReactDOM.render` inside a Computation.
    // In that case, we want to opt out of the normal behavior of nested
    // Computations, where if the outer one is invalidated or stopped,
    // it stops the inner one.
    tracker_1.Tracker.nonreactive(function () { return tracker_1.Tracker.autorun(function (c) {
        refs.computation = c;
        var data = reactiveFn(c);
        if (c.firstRun) {
            // Always run the reactiveFn on firstRun
            refs.trackerData = data;
        }
        else if (!skipUpdate || !skipUpdate(refs.trackerData, data)) {
            // For any reactive change, forceUpdate and let the next render rebuild the computation.
            forceUpdate();
        }
    }); });
    // To clean up side effects in render, stop the computation immediately
    if (!refs.isMounted) {
        meteor_1.Meteor.defer(function () {
            if (!refs.isMounted && refs.computation) {
                refs.computation.stop();
                delete refs.computation;
            }
        });
    }
    react_1.useEffect(function () {
        // Let subsequent renders know we are mounted (render is committed).
        refs.isMounted = true;
        // In some cases, the useEffect hook will run before Meteor.defer, such as
        // when React.lazy is used. In those cases, we might as well leave the
        // computation alone!
        if (!refs.computation) {
            // Render is committed, but we no longer have a computation. Invoke
            // forceUpdate and let the next render recreate the computation.
            if (!skipUpdate) {
                forceUpdate();
            }
            else {
                tracker_1.Tracker.nonreactive(function () { return tracker_1.Tracker.autorun(function (c) {
                    var data = reactiveFn(c);
                    refs.computation = c;
                    if (!skipUpdate(refs.trackerData, data)) {
                        // For any reactive change, forceUpdate and let the next render rebuild the computation.
                        forceUpdate();
                    }
                }); });
            }
        }
        // stop the computation on unmount
        return function () {
            var _a;
            (_a = refs.computation) === null || _a === void 0 ? void 0 : _a.stop();
            delete refs.computation;
            refs.isMounted = false;
        };
    }, []);
    return refs.trackerData;
};
var useTrackerWithDeps = function (reactiveFn, deps, skipUpdate) {
    if (skipUpdate === void 0) { skipUpdate = null; }
    var forceUpdate = useForceUpdate();
    var refs = react_1.useRef({ reactiveFn: reactiveFn }).current;
    // keep reactiveFn ref fresh
    refs.reactiveFn = reactiveFn;
    react_1.useMemo(function () {
        // To jive with the lifecycle interplay between Tracker/Subscribe, run the
        // reactive function in a computation, then stop it, to force flush cycle.
        var comp = tracker_1.Tracker.nonreactive(function () { return tracker_1.Tracker.autorun(function (c) {
            var data = refs.reactiveFn();
            if (c.firstRun) {
                refs.data = data;
            }
            else if (!skipUpdate || !skipUpdate(refs.data, data)) {
                refs.data = data;
                forceUpdate();
            }
        }); });
        // Stop the computation immediately to avoid creating side effects in render.
        // refers to this issues:
        // https://github.com/meteor/react-packages/issues/382
        // https://github.com/meteor/react-packages/issues/381
        if (refs.comp)
            refs.comp.stop();
        // In some cases, the useEffect hook will run before Meteor.defer, such as
        // when React.lazy is used. This will allow it to be stopped earlier in
        // useEffect if needed.
        refs.comp = comp;
        // To avoid creating side effects in render, stop the computation immediately
        meteor_1.Meteor.defer(function () {
            if (!refs.isMounted && refs.comp) {
                refs.comp.stop();
                delete refs.comp;
            }
        });
    }, deps);
    react_1.useEffect(function () {
        // Let subsequent renders know we are mounted (render is committed).
        refs.isMounted = true;
        if (!refs.comp) {
            refs.comp = tracker_1.Tracker.nonreactive(function () { return tracker_1.Tracker.autorun(function (c) {
                var data = refs.reactiveFn(c);
                if (!skipUpdate || !skipUpdate(refs.data, data)) {
                    refs.data = data;
                    forceUpdate();
                }
            }); });
        }
        return function () {
            refs.comp.stop();
            delete refs.comp;
            refs.isMounted = false;
        };
    }, deps);
    return refs.data;
};
function useTrackerClient(reactiveFn, deps, skipUpdate) {
    if (deps === void 0) { deps = null; }
    if (skipUpdate === void 0) { skipUpdate = null; }
    if (deps === null || deps === undefined || !Array.isArray(deps)) {
        if (typeof deps === "function") {
            skipUpdate = deps;
        }
        return useTrackerNoDeps(reactiveFn, skipUpdate);
    }
    else {
        return useTrackerWithDeps(reactiveFn, deps, skipUpdate);
    }
}
var useTrackerServer = function (reactiveFn) {
    return tracker_1.Tracker.nonreactive(reactiveFn);
};
// When rendering on the server, we don't want to use the Tracker.
// We only do the first rendering on the server so we can get the data right away
var _useTracker = meteor_1.Meteor.isServer
    ? useTrackerServer
    : useTrackerClient;
function useTrackerDev(reactiveFn, deps, skipUpdate) {
    if (deps === void 0) { deps = null; }
    if (skipUpdate === void 0) { skipUpdate = null; }
    function warn(expects, pos, arg, type) {
        console.warn("Warning: useTracker expected a " + expects + " in it's " + pos + " argument "
            + ("(" + arg + "), but got type of `" + type + "`."));
    }
    if (typeof reactiveFn !== 'function') {
        warn("function", "1st", "reactiveFn", reactiveFn);
    }
    if (deps && skipUpdate && !Array.isArray(deps) && typeof skipUpdate === "function") {
        warn("array & function", "2nd and 3rd", "deps, skipUpdate", typeof deps + " & " + typeof skipUpdate);
    }
    else {
        if (deps && !Array.isArray(deps) && typeof deps !== "function") {
            warn("array or function", "2nd", "deps or skipUpdate", typeof deps);
        }
        if (skipUpdate && typeof skipUpdate !== "function") {
            warn("function", "3rd", "skipUpdate", typeof skipUpdate);
        }
    }
    var data = _useTracker(reactiveFn, deps, skipUpdate);
    checkCursor(data);
    return data;
}
exports.useTracker = meteor_1.Meteor.isDevelopment
    ? useTrackerDev
    : _useTracker;
